""" Core utility functions to work with Jupyter Notebooks. """
#pylint: disable=import-outside-toplevel
import os
import re
import json
import warnings



class StringWithDisabledRepr(str):
    """ String with disabled repr. Used to avoid cluttering repr from function outputs. """
    def __repr__(self):
        """ Shorten the repr of a string. """
        return f'<StringWithDisabledRepr at {hex(id(self))}. Use `str`/`print` explicitly!>'



def in_notebook():
    """ Return True if in Jupyter notebook and False otherwise. """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True
        if shell == 'TerminalInteractiveShell':
            return False
        return False
    except NameError:
        return False

def get_notebook_path():
    """ Return the full absolute path of the current Jupyter notebook,
    for example, `/path/path/path/My_notebook_title.ipynb`.

    If run outside Jupyter notebook, returns None.
    """
    #pylint: disable=missing-timeout
    if not in_notebook():
        return None

    import requests
    import ipykernel

    # Id of the current running kernel: a string uid
    kernel_id = re.search('kernel-(.*).json', ipykernel.connect.get_connection_file()).group(1)

    # Get running servers for both JupyterLab v2.# and v3.#
    from notebook.notebookapp import list_running_servers as list_running_servers_v2
    from jupyter_server.serverapp import list_running_servers as list_running_servers_v3
    servers = list(list_running_servers_v2()) + list(list_running_servers_v3())

    for server in servers:
        root_dir = server.get('root_dir') or server.get('notebook_dir')
        response = requests.get(requests.compat.urljoin(server['url'], 'api/sessions'),
                                params={'token': server.get('token', '')})

        for params in json.loads(response.text):
            if params['kernel']['id'] == kernel_id:
                relative_path = params['notebook']['path']
                return os.path.join(root_dir, relative_path)
    raise ValueError(f'Unable to find kernel `{kernel_id}` in {len(servers)} servers!')

def get_notebook_name():
    """ Return the title of the current Jupyter notebook without base directory and extension,
    for example, `My_notebook_title`.

    If run outside Jupyter notebook, returns None.
    """
    if not in_notebook():
        return None

    return os.path.splitext(get_notebook_path())[0].split('/')[-1]


def notebook_to_script(path_script, path_notebook=None, ignore_markdown=True, return_info=False):
    """ Convert a notebook to a script. """
    import nbformat #pylint: disable=import-outside-toplevel
    path_notebook = path_notebook or get_notebook_path()
    if path_notebook is None:
        raise ValueError('Provide path to Jupyter Notebook or run `notebook_to_script` inside of it!')

    # Read notebook as list of cells
    notebook = nbformat.read(path_notebook, as_version=4)

    code_lines = []
    cell_number = 1
    cell_line_numbers = {}

    for cell in notebook['cells']:
        if ignore_markdown and cell['cell_type'] != 'code':
            continue

        cell_lines = cell['source'].split('\n')
        cell_lines.insert(0, f'\n### [{cell_number}] cell')

        # Comment cell/line magics
        for j, line in enumerate(cell_lines):
            if line.startswith('%') or line.startswith('!'):
                cell_lines[j] = '### ' + line

        code_line_number = len(code_lines) + 1
        cell_line_numbers[cell_number] = range(code_line_number, code_line_number + len(cell_lines))

        code_lines.extend([line.strip('\n') for line in cell_lines])
        code_lines.append('')
        cell_number += 1

    code = '\n'.join(code_lines).strip()

    with open(path_script, 'w', encoding='utf-8') as file:
        file.write(code)

    if return_info:
        return {'code': StringWithDisabledRepr(code),
                'cell_line_numbers': cell_line_numbers}
    return None



def get_available_gpus(n=1, min_free_memory=0.9, max_processes=2, verbose=False, raise_error=False):
    """ Select `n` gpus from available and free devices.

    Parameters
    ----------
    n : int, str
        If `max`, then use maximum number of available devices.
        If int, then number of devices to select.
    min_free_memory : float
        Minimum percentage of free memory on a device to consider it free.
    max_processes : int
        Maximum amount of computed processes on a device to consider it free.
    verbose : bool
        Whether to show individual device information.
    raise_error : bool
        Whether to raise an exception if not enough devices are available.

    Returns
    -------
    List with indices of available GPUs
    """
    try:
        import nvidia_smi
    except ImportError as exception:
        raise ImportError('Install Python interface for nvidia_smi') from exception

    nvidia_smi.nvmlInit()
    n_devices = nvidia_smi.nvmlDeviceGetCount()

    available_devices, memory_usage = [], []
    for i in range(n_devices):
        handle = nvidia_smi.nvmlDeviceGetHandleByIndex(i)
        info = nvidia_smi.nvmlDeviceGetMemoryInfo(handle)

        fraction_free = info.free / info.total
        num_processes = len(nvidia_smi.nvmlDeviceGetComputeRunningProcesses(handle))

        consider_available = (fraction_free > min_free_memory) & (num_processes <= max_processes)
        if consider_available:
            available_devices.append(i)
            memory_usage.append(fraction_free)

        if verbose:
            print(f'Device {i} | Free memory: {fraction_free:4.2f} | '
                  f'Number of running processes: {num_processes:>2} | Free: {consider_available}')

    if isinstance(n, str) and n.startswith('max'):
        n = len(available_devices)

    if len(available_devices) < n:
        msg = f'Not enough free devices: requested {n}, found {len(available_devices)}'
        if raise_error:
            raise ValueError(msg)
        warnings.warn(msg, RuntimeWarning)

    # Argsort of `memory_usage` in a descending order
    indices = sorted(range(len(available_devices)), key=memory_usage.__getitem__, reverse=True)
    available_devices = [available_devices[i] for i in indices]
    return sorted(available_devices[:n])

def get_gpu_free_memory(index):
    """ Get free memory of a device. """
    try:
        import nvidia_smi
    except ImportError as exception:
        raise ImportError('Install Python interface for nvidia_smi') from exception

    nvidia_smi.nvmlInit()
    nvidia_smi.nvmlDeviceGetCount()
    handle = nvidia_smi.nvmlDeviceGetHandleByIndex(index)
    info = nvidia_smi.nvmlDeviceGetMemoryInfo(handle)

    return info.free / info.total

def set_gpus(n=1, min_free_memory=0.9, max_processes=2, verbose=False, raise_error=False):
    """ Set the `CUDA_VISIBLE_DEVICES` variable to `n` available devices.

    Parameters
    ----------
    n : int, str
        If `max`, then use maximum number of available devices.
        If int, then number of devices to select.
    min_free_memory : float
        Minimum percentage of free memory on a device to consider it free.
    max_processes : int
        Maximum amount of computed processes on a device to consider it free.
    verbose : bool or int
        Whether to show individual device information.
        If 0 or False, then no information is displayed.
        If 1 or True, then display the value assigned to `CUDA_VISIBLE_DEVICES` variable.
        If 2, then display memory and process information for each device.
    raise_error : bool
        Whether to raise an exception if not enough devices are available.
    """
    #pylint: disable=consider-iterating-dictionary
    if 'CUDA_VISIBLE_DEVICES' in os.environ.keys():
        str_devices = os.environ["CUDA_VISIBLE_DEVICES"]
        warnings.warn(f'`CUDA_VISIBLE_DEVICES` is already set to "{str_devices}"!')
        return [int(d) for d in str_devices.split(',')]

    devices = get_available_gpus(n=n, min_free_memory=min_free_memory, max_processes=max_processes,
                                 verbose=(verbose==2), raise_error=raise_error)
    str_devices = ','.join(str(i) for i in devices)
    os.environ['CUDA_VISIBLE_DEVICES'] = str_devices

    newline = "\n" if verbose==2 else ""
    if verbose:
        print(f'{newline}`CUDA_VISIBLE_DEVICES` set to "{str_devices}"')
    return devices



""" Resource -- a class to describe a property of an entry. Refer to class documentation for more. """
from enum import Enum, auto


class Resource(Enum):
    """ Enumeration of all possible properties of an entry.
    An entry is any monitored entity -- Python process, Jupyter Notebook or device.

    A good example of property of a Python process is the amount of CPU resources it uses,
    or the number of its subprocesses, or the amount of GPUs (or other accelerators) taken by this process.
    At the same time, we consider less obvious properties to be a Resource: the name of process, its disk path and
    even type (process, notebook or device).
    Moreover, we consider table elements and prettifyings to be Resources as well: that help us to use the
    same notation for both building and formatting the resulting table.

    A sequence of Resources supposed to define both resources to fetch from system sensors and structure of the
    table (used columns and delimiters): we use `ResourceFormatter` class for that purpose.

    A dictionary of all properties with their values to describe an entry is `ResourceEntry`.
    Entries are collected into `ResourceTables`, which provide methods to beautifully format their data.
    Refer to those classes for continuation of the usage logic of Resources.

    Using the enumeration makes sure that the same named unique constants are used throughout the code base.
    Nevetheless, `parse_alias` method makes sure that we can convert string like `'device_util'` to a correct Resource.
    It comes handy when we get strings from command line, and we use it to process every key that can come from user.

    Members of this enumeration have prefixes to distinguish their supposed use:
        - `` to describe Python processes and Jupyter Notebooks properties like name, path, PID
        - `DEVICE_` to describe properties of accelerators like memory taken, temperature, and utilization
        - `TABLE_` to denote elements that are parts of the formatted table.

    As all `Enum`s, does not work well with `autoreload` Jupyter magic.


    TODO: a potential improvement of the logic is to use values not only for aliases, but to point out the
    resources to collect for complex columns, which are not fetched themselves, but rely on other collected resources.
    For example, the `DEVICE_SHORT_ID` requires the `DEVICE_ID` and we can define it here instead of
    the corresponding if-clause in the `ResourceInspector.get_device_table`.
    TODO: add encoder/decoder utilization
    """
    # Possible columns in python-table: notebooks and python scripts
    PROCESS = auto()
    TYPE = auto()
    NAME = 'process_name'
    PATH = auto()
    CMDLINE = auto()
    PID = auto()
    PPID = auto()
    NGID = auto()
    HOST_PID = auto()
    PYTHON_PPID = auto()
    CREATE_TIME = auto()
    KERNEL = 'kernel_id'
    STATUS = auto()

    RSS = auto()
    CPU = auto()

    # Possible columns in device-table: information about each device
    DEVICE_ID = auto()
    DEVICE_NAME = auto()
    DEVICE_MEMORY_UTIL = auto()
    DEVICE_MEMORY_USED = 'memory'
    DEVICE_MEMORY_TOTAL = auto()
    DEVICE_TEMP = ['temp', 'temperature']
    DEVICE_POWER_USED = 'power'
    DEVICE_POWER_TOTAL = auto()
    DEVICE_FAN = 'fan'
    DEVICE_UTIL = 'util'
    DEVICE_UTIL_ENC = auto()
    DEVICE_UTIL_DEC = auto()

    # Aggregated process info
    DEVICE_PROCESS_N = auto()
    DEVICE_PROCESS_PID = 'device_pid'
    DEVICE_PROCESS_MEMORY_USED = 'process_memory'

    # Used for better repr in formatter tables
    TABLE_DELIMITER1 = auto()
    TABLE_DELIMITER2 = auto()
    DEVICE_SHORT_ID = 'short_id'

    USES_DEVICE = auto()
    IS_PARENT = auto()

    # Moving averages
    DEVICE_UTIL_MA = 'util_ma'


    def __repr__(self):
        return self.name

    @staticmethod
    def parse_alias(alias):
        """ Convert a string `alias` into member of the Resource enumeration. """
        if isinstance(alias, str):
            alias = alias.lower()
            if alias in Resource.ALIAS_TO_RESOURCE:
                alias = Resource.ALIAS_TO_RESOURCE[alias]
        return alias

    def to_format_data(self, terminal, **kwargs):
        """ Create a string template and data for a given `self=resource`.
        For more information about formatting refer to `ResourceTable.format` method.
        """
        _ = kwargs
        style, string = None, None

        # Process description
        if self == Resource.NAME:
            style = terminal.bold
            string = 'PROCESS NAME'
        elif self in [Resource.TYPE, Resource.PID, Resource.PYTHON_PPID,
                          Resource.STATUS, Resource.CREATE_TIME, Resource.KERNEL]:
            style = terminal.normal

        # Process resources
        elif self in [Resource.RSS, Resource.CPU]:
            style = terminal.bold + terminal.cyan

        # Device description
        elif self == Resource.DEVICE_ID:
            style = terminal.blue
            string = 'DEVICE NAME ' + terminal.cyan + '[ID]'
        elif self == Resource.DEVICE_SHORT_ID:
            style = terminal.blue
            string = 'DEVICE ID'
        elif self == Resource.DEVICE_PROCESS_PID:
            string = 'DEVICE PID'

        # Device resources
        elif self == Resource.DEVICE_MEMORY_USED:
            style = terminal.yellow
        elif self == Resource.DEVICE_PROCESS_MEMORY_USED:
            style = terminal.yellow
        elif self == Resource.DEVICE_POWER_USED:
            style = terminal.magenta
        elif self in [Resource.DEVICE_UTIL]:
            style = terminal.green
            string = 'UTIL'
        elif self in [Resource.DEVICE_UTIL_MA]:
            style = terminal.green
            string = 'UTIL AVERAGE'
        elif self == Resource.DEVICE_TEMP:
            style = terminal.red

        # Table elements
        elif self == Resource.TABLE_DELIMITER1:
            string = '┃'
        elif self == Resource.TABLE_DELIMITER2:
            string = '║'

        # Default values
        if style is None:
            style = ''
        if string is None:
            string = self.name.replace('', '').replace('DEVICE_', '').replace('_USED', '').replace('_', ' ')
        return style, string


# Dictionary with aliases for each Resource: more aliases can be added by setting values in Enum instead of `auto`
# Added to the class attributes after its creation so it is not a member of actual enumeration.
ALIAS_TO_RESOURCE, RESOURCE_TO_ALIAS = {}, {}
for resource in Resource.__members__.values():
    name, aliases = resource.name, resource.value
    aliases_from_name = [name, name.lower()]

    aliases = aliases if isinstance(aliases, list) else [aliases]
    aliases = [alias for alias in aliases if not isinstance(alias, int)]
    aliases = aliases_from_name + aliases

    ALIAS_TO_RESOURCE.update({alias : resource for alias in aliases})
    RESOURCE_TO_ALIAS[resource] = [alias for alias in aliases if isinstance(alias, str)][-1]
Resource.ALIAS_TO_RESOURCE = ALIAS_TO_RESOURCE
Resource.RESOURCE_TO_ALIAS = RESOURCE_TO_ALIAS



""" ResourceEntry -- a dict-like class to hold all properties (Resources) of an entry.
ResourceTable -- sequence of multiple ResourceEntries with interface of merging, updating, sorting for multiple tables.
Can be formatted into beautiful colored string representation by using `format` method.
"""

from .resource import Resource
from .resource_entry import ResourceEntry

class ResourceTable:
    """ Container for multiple ResourceEntries.
    Provides API to work with multiple tables, collected about different types of entries. In this module, we use:
        - a table about all Python processes, collected with `psutil` and the likes
        - a table about Jupyter Notebooks, collected with requests to `jupyter_server` and `notebook` interfaces
        - a table about system devices, collected with `nvidia-smi` Python bindings.

    Creating a final table for actual representation requires some merges, sorts, filtering, etc.
    For convenience, we implement all of the usual table methods: `merge`, `update`, `sort`, `filter`, as well as their
    versions for working with tables with defined `index`.

    For the most part, this class acts as a lightweight version of a`Pandas.DataFrame`: it saves us a huge dependency
    and also allows to tweak some of the metods to our needs. As our tables are tiny, efficiency is not a concern.
    If the `pandas` is already installed, `ResourceTable` can be converted to dataframe by using `to_pd` method.

    Also provides method `format` for table visualization as a beautiful colored string representation.
    Most of the design choices in this module are dictated by the needs of this method: classes are made as they are
    so it is easy to use them there.

    Current implementation uses `list` as the underlying container, which is appended by `ResourceEntries`.
    Checks that the set of appended columns is the same as columns in previous entries, but does not guarantee
    consistency of columns if initialized from list.

    # TODO: can require `index` and/or `columns` at table creation, should we want so, for consistency
    # TODO: can create a `PandasResourceTable` with the same API, but `pandas.DataFrame` under the hood
    """
    #pylint: disable=self-cls-assignment
    def __init__(self, data=None):
        self._data = [] if data is None else [ResourceEntry(entry) for entry in data]

    @property
    def data(self):
        """ Property to make sure that every entry is an instance of `ResourceEntry`. """
        return self._data

    @data.setter
    def data(self, value):
        """ Property to make sure that every entry is an instance of `ResourceEntry`. """
        self._data = [ResourceEntry(entry) for entry in value]


    # Basic inner workings
    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        """ Multiple ways to get items from the table:
            - integer to get i-th entry from the table. Returns an instance of `ResourceEntry`
            - slice to get range of entries from the table. Returns an instance of `ResourceTable`
            - member of Resource enumeration or string alias to get column values. Returns list
            - list with bools of the same length, as the self to signal which entries to include.
            Returns an instance of `ResourceTable`
        """
        if isinstance(key, int):
            return self.data[key]

        if isinstance(key, slice):
            return ResourceTable(self.data[key])

        if isinstance(key, (str, Resource)):
            key = Resource.parse_alias(key)
            return [entry[key] for entry in self]

        # Iterable with bools
        if len(key) == len(self):
            return ResourceTable(data=[entry for entry, flag in zip(self, key) if flag])

        raise TypeError(f'Unsupported type {type(key)} for getitem!')

    @property
    def columns(self):
        """ List of present columns. """
        return None if len(self.data) == 0 else list(self.data[0].keys())

    def append(self, entry):
        """ Check if the `keys` of `entry` match columns of the table. Wrap with `ResourceEntry`, if needed. """
        entry_columns = list(entry.keys())

        if self.columns is not None:
            if set(self.columns) != set(entry_columns):
                raise ValueError('Trying to append entry with different set of columns!')

        entry = ResourceEntry(entry)
        self.data.append(entry)

    def maybe_copy(self, return_self):
        """ Inspired by Pandas. `return_self` coincides with `inplace` flag of the calling method. """
        return self if return_self else ResourceTable(self.data)


    # Logic
    def merge(self, other, self_key, other_key):
        """ Create a new `ResourceTable` with info merged from `self` and `other`,
        from entries where values of `self_key` and `other_key` columns match. All of the entries in `self` remain.
        If no matched entry present in `other`, values of additional columns are set to None.
        """
        if not other:
            return self

        # Assert
        if self_key not in self.columns:
            raise ValueError(f'Key `{self_key}` is not found in `self`!')
        if other_key not in other.columns:
            raise ValueError(f'Key `{other_key}` is not found in `other`!')

        for self_column in self.columns:
            for other_column in other.columns:
                if self_column == other_column:
                    raise ValueError(f'Column `{self_column}` present in both tables!')

        # Actual logic
        result = ResourceTable()
        template = {**{key : None for key in self.columns},
                    **{key : None for key in other.columns}}

        for self_entry in self:
            self_value = self_entry[self_key]
            flag = False

            for other_entry in other:
                other_value = other_entry[other_key]

                if self_value == other_value:
                    merged_entry = {**template, **self_entry, **other_entry}
                    result.append(merged_entry)
                    flag = True
            if flag is False:
                merged_entry = {**template, **self_entry}
                result.append(merged_entry)
        return result

    def update(self, other, self_key, other_key, inplace=True):
        """ Update `self` table with `other` table, matching entries on `self_key` and `other_key` columns.
        All of the columns in `other` should be present in `self`. All of the entries in `self` remain.
        """
        self = self.maybe_copy(return_self=inplace)

        # Assert
        if self_key not in self.columns:
            raise ValueError(f'Key `{self_key}` is not found in `self`!')
        if other_key not in other.columns:
            raise ValueError(f'Key `{other_key}` is not found in `other`!')

        if set(other.columns).intersection(set(self.columns)) != set(other.columns):
            difference = set(other.columns).difference(set(other.columns).intersection(set(self.columns)))
            raise ValueError(f'Columns of `other` should be a strict subset of `self` columns! Excess: {difference}')

        # Actual logic
        for self_entry in self:
            self_value = self_entry[self_key]
            for other_entry in other:
                if self_value == other_entry[other_key]:
                    self_entry.update(other_entry)
        return self

    def unroll(self, inplace=True):
        """ Unroll entries with sequence-values into multiple separate entries. """
        self = self.maybe_copy(return_self=inplace)

        unrolled = []
        for entry in self:
            lens = []
            for value in entry.values():
                if isinstance(value, list):
                    lens.append(len(value))

            if len(set(lens)) != 1:
                raise ValueError('Entry items have different lengths!')
            n = lens[0]

            if n == 0:
                new_entry = {key : (value if not isinstance(value, list) else None)
                            for key, value in entry.items()}
                unrolled.append(new_entry)
            else:
                for i in range(n):
                    new_entry = {key : (value if not isinstance(value, list) else value[i])
                                for key, value in entry.items()}
                    unrolled.append(new_entry)
        self.data = unrolled
        return self

    def sort(self, key, default=0.0, reverse=False, inplace=True):
        """ Sort the table based on `key`.

        Parameters
        ----------
        key : Resource, string or sequence of them
            Keys to sort on.
        default : number or sequence of numbers
            Default values to use instead of Nones.
        reverse : bool or sequence of bools
            Whether to use descending or ascending order of sort, individual for each key.
        inplace : bool
            Whether to return the same instance with changed data or a completely new instance.
        """
        self = self.maybe_copy(return_self=inplace)

        # Parse parameters into the similar list structure
        key = key if isinstance(key, (tuple, list)) else [key]
        default = default if isinstance(default, (tuple, list)) and len(default) == len(key) else [default] * len(key)
        reverse = reverse if isinstance(reverse, (tuple, list)) and len(reverse) == len(key) else [reverse] * len(key)
        def itemgetter(entry):
            result = []
            for key_, default_, reverse_ in zip(key, default, reverse):
                key_ = Resource.parse_alias(key_)
                sign = -1 if reverse_ is True else +1
                value = entry.get(key_, None)
                value = value if value is not None else default_
                result.append(sign * value)
            return tuple(result)

        self.data.sort(key=itemgetter)
        return self

    def filter(self, condition, inplace=True):
        """ Filter entries, based on `condition`. Keep only those which evaluate to True. """
        self = self.maybe_copy(return_self=inplace)
        self.data = [entry for entry in self if condition(entry)]
        return self


    # Work with index
    def set_index(self, index, inplace=True):
        """ Select a column as the index of the table. Stable sorts on unique values of the index.
        Under the hood, creates a list of unique values of chosen column to use in later methods.
        """
        self = self.maybe_copy(return_self=inplace)
        index_values = []
        for entry in self:
            value = entry[index]
            if value not in index_values:
                index_values.append(value)

        data = []
        for value in index_values:
            for entry in self:
                if entry[index] == value:
                    data.append(entry)

        self.data = data
        self.index = index
        self.index_values = index_values
        return self

    def _extract_subtable(self, index_value):
        """ Extract subtable, corresponding to one of the current index values. """
        subtable = ResourceTable()
        for entry in self:
            if entry[self.index] == index_value:
                subtable.append(entry)
        return subtable

    def split_by_index(self):
        """ Split the table into a list of subtables, corresponding to unique index values. """
        subtables = []
        for index_value in self.index_values:
            subtable = self._extract_subtable(index_value)
            subtables.append(subtable)

        return subtables

    def sort_by_index(self, key, default=0.0, reverse=False, inplace=True, aggregation=max):
        """ Sort unique index values, based on aggregated values of `key` columns in their subtables.
        Has the same semantics, as `sort` method.
        """
        self = self.maybe_copy(return_self=inplace)

        # Parse parameters into the similar list structure
        key = key if isinstance(key, (tuple, list)) else [key]
        default = default if isinstance(default, (tuple, list)) else [default] * len(key)
        reverse = reverse if isinstance(reverse, (tuple, list)) else [reverse] * len(key)
        aggregation = aggregation if isinstance(aggregation, (tuple, list)) else [aggregation] * len(key)

        def itemgetter(index_value):
            subtable = self._extract_subtable(index_value)
            result = []

            for key_, default_, reverse_, aggregation_ in zip(key, default, reverse, aggregation):
                sign = -1 if reverse_ is True else +1
                value = subtable.aggregate(key=key_, default=default_, aggregation=aggregation_)
                result.append(sign * value)
            return tuple(result)

        self.index_values.sort(key=itemgetter)
        return self

    def filter_by_index(self, condition, inplace=True):
        """ Filter entries in subtables, based on `condition`. Keep only those which evaluate to True.
        Essentially, the same as `filter`, by also updates `index_values` list.
        """
        self = self.maybe_copy(return_self=inplace)

        data = []
        index_values = []
        for index_value, subtable in zip(self.index_values, self.split_by_index()):
            filtered_subtable = subtable.filter(condition=condition, inplace=False)
            if len(filtered_subtable) > 0:
                data.extend(subtable.data)
                index_values.append(index_value)

        self.data = data
        self.index_values = index_values
        return self

    def filter_on_index(self, condition, inplace=True):
        """ Filter subtables and index values, based on `condition`, evaluated on them.
        Keep only those which evaluate to True.
        """
        self = self.maybe_copy(return_self=inplace)

        data = []
        index_values = []
        for index_value, subtable in zip(self.index_values, self.split_by_index()):
            if condition(index_value, subtable):
                data.extend(subtable.data)
                index_values.append(index_value)

        self.data = data
        self.index_values = index_values
        return self

    def aggregate_by_index(self, key, default=0.0, aggregation=max):
        """ Get aggregates of `key` column for each index value. """
        result = []
        for subtable in self.split_by_index():
            value = subtable.aggregate(key=key, default=default, aggregation=aggregation)
            result.append(value)
        return result


    # Work with columns
    def apply(self, key, function):
        """ Apply `function` to a `key` column. """
        for entry in self:
            entry[key] = function(entry[key])

    def add_column(self, key, function):
        """ Apply `function` to each entry in the table and store as `key` column. """
        for entry in self:
            entry[key] = function(entry)

    def aggregate(self, key, default=0.0, aggregation=max):
        """ Apply `aggregation` to values from a `key`-column. `None` values are changed to `default`. """
        data = [(item if item is not None else default) for item in self[key]]
        return aggregation(data)


    # Display
    def __str__(self):
        return repr('\n'.join([str(entry) for entry in self.data]))

    def to_format_data(self, resource, terminal, **kwargs):
        """ Create a string template and data for a given `resource` column from entire table.
        Works by aggregation information in the `resource` column and making string from it.
        For more information about formatting refer to `ResourceTable.format` method.

        TODO: not used currently, here to show the idea
        """
        style, string = None, None

        if resource == Resource.DEVICE_SHORT_ID:
            data = self.aggregate(Resource.DEVICE_ID, default=None, aggregation=list)
            data = [str(item) for item in sorted(set(data)) if item is not None]

            style = ''
            string = f'[{", ".join(data) or "-"}]'
        else:
            # Works for most float-resources: sum over all elements of the table
            data = self.aggregate(resource, default=0.0, aggregation=sum)
            entry = ResourceEntry({resource : data})
            string, style = entry.to_format_data(resource, terminal, **kwargs)

        return string, style


    def format(self, terminal, formatter, hide_similar=True,
               add_header=True, underline_header=True, bold_header=True, separate_header=True, separate_index=True,
               process_memory_format='GB', device_memory_format='MB'):
        """ Create a colored textual representation of a table.

        For each element of `formatter`, that has the `include` flag set to True, we create a column.
        Each column consists of header (optional), body (info from the actual table entries), and separators (optional).
            - each column represents a Resource, and by knowing it we can compute the `style` and `name` of this column.
            That is done by the `Resource.to_format_data` method. For most resources, `style` is a color or
            text formatting, and it is enough to format the string from entries.
            The header string is produced by (`style` + `name`).

            - then we iterate over entries in the table. Each entry is requested to create a `style` and `string` for
            current resource: that is done by the `ResourceEntry.to_format_data` method.
            Note that we can't use just the resource and its value from the entry to create the `string`, as some of the
            columns require multiple items: for example, device memory requires current and total values.
            `style`, returned by the entry, can completely override the style from the Resource, but for the most cases,
            it is empty or just adds bold/underlines.
            The overall string for the entry is produced by (`main_style` + `entry_style` + `string`).

        After getting all the data for a column, we justify it so all the elements have the same width.
        This process is complicated by the fact that most of the strings are colored / formatted, so we use custom
        functions instead of `str` methods. Finally, we add separators to the table, if needed.

        We rely on `formatter` for getting the structure of a table instead of looking at the present columns,
        as some of the table elements are more complicated than just retrieving the resource value and formatting it.

        Parameters
        ----------
        terminal : blessed.Terminal
            Terminal to use for colors and text formatting symbols.
        formatter : ResourceFormatter
            Formatter to get the structure of a table, as well as additional display parameters for each of them.
        hide_similar : bool
            Whether to allow some of the elements to be hidden.
            For example, the name of the process in all rows after the first.
        add_header, underline_header, bold_header, separate_header : bool
            Parameters of table header.
        separate_index : bool
            Whether to separate indexed subtables with table separators.
        process_memory_format, device_memory_format : {'KB', 'MB', 'GB}
            Memory unit to use for representation process RSS / device memory taken.

        TODO: a potential improvement is to use bold_normal only for table elements.
        TODO: a potential improvement is to add `limit` parameter to show no more lines for each
        index value than requested, ending the subtable representation with aggregation of remaining processes.
        All of the necessary functions are already implemented: all we need to do is to add more cases to the switch in
        `ResourceTable.to_format_data` and call it here.
        TODO: a potential improvement is to add capability to handle multi-line strings for individual resources:
        that would require transposing the loop of lines creation, but overall not that hard.
        """
        subtables = self.split_by_index()
        kwargs = {'process_memory_format' : process_memory_format,
                  'device_memory_format' : device_memory_format}

        lines = [[] for _ in range(1 + len(self))]
        for column_dict in formatter.included_only:
            # Retrieve parameters of the column display
            resource = column_dict['resource']
            hidable, min_width = column_dict.get('hidable', False), column_dict.get('min_width', 0)
            column_kwargs = {**kwargs,
                             **{key : value for key, value in column_dict.items()
                                if key != 'resource'}}

            styles, strings = [], []

            # Table header: names of the columns
            main_style, header_string = resource.to_format_data(terminal=terminal, **column_kwargs)
            if add_header:
                header_style = ''
                if 'TABLE_DELIMITER' not in resource.name:
                    header_style += (terminal.underline if underline_header else '')
                    header_style += (terminal.bold if bold_header else '')
                styles.append(header_style)
                strings.append(header_string)

            # Body of the table: add sublines for each table entry
            for subtable in subtables:
                for i, entry in enumerate(subtable):
                    style, string = entry.to_format_data(resource=resource, terminal=terminal, **column_kwargs)

                    # Changes based on the position of entry
                    if hide_similar and hidable and i > 0:
                        style, string = '', ''

                    if False and i: # pylint: disable=condition-evals-to-constant
                        # TODO: aggregate info by using `aggregate` method
                        # for a given resource and create a ResourceEntry out of it to make style/string
                        remaining_table = subtable[1:]
                        style, string = remaining_table.to_format_data(resource=resource, terminal=terminal,
                                                                       **column_kwargs)

                    styles.append(style)
                    strings.append(string)

            # Modify header style: if any entry used bold, use it in the header as well
            if add_header:
                if len(styles) > 1:
                    styles[0] += terminal.bold * max(terminal.bold in style for style in styles)

            # Make every string the same width
            strings = [main_style + style + string + terminal.normal
                       for style, string in zip(styles, strings)]
            max_len = max(terminal.length(string) for string in strings)
            width = max(min_width, max_len)
            strings = [terminal.rjust(string, width) for string in strings]

            for line, string in zip(lines, strings):
                line.append(string)

        lines = [' '.join(line).rstrip() + terminal.normal for line in lines]

        # Add separators between index values
        if separate_index or (add_header and separate_header):
            s, separator_indices = 1 if add_header else 0, []
            for subtable in subtables:
                separator_indices.append(s)
                s += len(subtable)

            if (add_header and separate_header) and separate_index:
                separator_indices = separator_indices[::-1]
            elif add_header and separate_header:
                separator_indices = separator_indices[:1]
            else:
                separator_indices = separator_indices[1:][::-1]

            separator = terminal.separator_symbol * terminal.length(lines[0])
            for idx in separator_indices:
                lines.insert(idx, separator)

        return lines


    # Pandas compatibility
    def to_pd(self):
        """ Convert the table to a `pandas.DataFrame`. """
        import pandas as pd #pylint: disable=import-outside-toplevel
        return pd.DataFrame(self.data)



""" Utility functions. """
#pylint: disable=redefined-builtin
import re
import platform
import linecache

import psutil



# Faster versions of some of `psutil` commands, available only on Linux
SYSTEM = platform.system()

def pid_to_name_generic(pid):
    """ Get `name` of a process by its PID. Platform-agnostic. """
    try:
        process = psutil.Process(pid)
        name = process.name()
    except (psutil.AccessDenied, psutil.ZombieProcess, psutil.NoSuchProcess):
        name = ''
    return name

def pid_to_name_linux(pid):
    """ Get `name` of a process by its PID on Linux. ~20% speed-up, compared to the `generic` version. """
    try:
        line = linecache.getline(f'/proc/{pid}/status', 1)
        name = line.strip().split()[1]
    except Exception: #pylint: disable=broad-except
        name = ''
    return name

pid_to_name = pid_to_name_linux if SYSTEM == 'Linux' else pid_to_name_generic


def pid_to_ngid_generic(pid):
    """ Get NGID of a process by its PID. """
    return pid

def pid_to_ngid_linux(pid):
    """ Get NGID of a process by its PID on Linux. Used as the PID on host for a process inside a container. """
    try:
        line = linecache.getline(f'/proc/{pid}/status', 5)
        ngid = line.strip().split()[1]
        ngid = int(ngid)
    except Exception: #pylint: disable=broad-except
        ngid = pid
    return ngid or pid

pid_to_ngid = pid_to_ngid_linux if SYSTEM == 'Linux' else pid_to_ngid_generic



# Utilities to work with strings containing terminal sequences
COLOR_REPLACER = re.compile(r"\x1b\[[;\d]*[A-Za-z]").sub
def true_len(string):
    """ Compute printable length of the string, ignoring terminal symbols like color / formatters / system signals. """
    length = len(COLOR_REPLACER('', string))
    for symbol in '％℃':
        length += string.count(symbol)
    return length

def true_rjust(string, width, fillchar=' '):
    """ Justify the string to the right, using printable length as the width. """
    return fillchar * (width - true_len(string)) + string

def true_center(string, width, fillchar=' '):
    """ Justify the string to the center, using printable length as the width. """
    true_pad = width - true_len(string)
    left = true_pad // 2
    right = true_pad - left
    return fillchar * left + string + fillchar * right



class FiniteList(list):
    """ List with finite number of elements: if the size is more """
    def __init__(self, *args, size=10, **kwargs):
        self.size = size
        super().__init__(*args, **kwargs)

    def append(self, obj):
        """ Append to the list. If length is bigger than allowed, remove the first element. """
        if len(self) >= self.size:
            self.pop(0)
        super().append(obj)

    def get_average(self, size=None):
        """ Compute average value of the last `size` elements.
        Returns `None`, if there are less than `size // 2` elements in the list.
        """
        size = size or self.size
        if len(self) > size // 2:
            sublist = self[-size:]
            return round(sum(sublist) / len(sublist))
        return None



def format_memory(number, format=3):
    """ Format memory in bytes to a desired format level. """
    level_to_unit = {1 : 'KB', 2 : 'MB', 3 : 'GB'}
    unit_to_level = {value : key for key, value in level_to_unit.items()}

    if isinstance(format, int):
        level, unit = format, level_to_unit[format]
    elif isinstance(format, str):
        level, unit = unit_to_level[format], format

    rounded = round(number / (1024 ** level), 1 if level>=3 else None)
    return rounded, unit



""" Controlling class for fetching tables with system information, merging them into view and displaying in stream. """
import os
import re
import json
import time

import psutil
import requests
from blessed import Terminal

import nvidia_smi

from .resource import Resource
from .resource_table import ResourceTable
from .utils import format_memory, pid_to_name, pid_to_ngid, true_len, true_rjust, true_center, FiniteList



KERNEL_ID_SEARCHER   = re.compile('kernel-(.*).json').search
VSCODE_KEY_SEARCHER  = re.compile('key=b"(.*)"').search
SCRIPT_NAME_SEARCHER = re.compile('python.* (.*).py').search


class ResourceInspector:
    """ A class to controll the process of gathering information about system resources into ResourceTables,
    merging them into views, and formatting into nice colored strings.

    TODO: correct working with VSCode Jupyter Notebooks
    TODO: make sure that everything works without sudo
    TODO: add more fallbacks for unavailable resources
    TODO: can add explicit __delete__ to call nvidia_smi.nvmlShutdown(), if we ever have problems with that
    """
    def __init__(self, formatter=None):
        self.formatter = formatter

        self._device_handles = None
        self._device_utils = None
        self.warnings = {}

    @property
    def device_handles(self):
        """ Cached handles of NVIDIA devices. """
        if self._device_handles is None:
            nvidia_smi.nvmlInit()
            n_devices = nvidia_smi.nvmlDeviceGetCount()

            self._device_handles = {device_id : nvidia_smi.nvmlDeviceGetHandleByIndex(device_id)
                                    for device_id in range(n_devices)}
        return self._device_handles

    @property
    def device_utils(self):
        """ Values of device utilization over time. """
        if self._device_utils is None:
            self._device_utils = {device_id : FiniteList(size=1000)
                                  for device_id in self.device_handles}
        return self._device_utils


    # Collect system resources into ResourceTables
    def get_device_table(self, formatter=None, window=20):
        """ Collect data about current device usage into two tables:
        one is indexed by device, the second is indexed by process on a device.

        Each value is collected only if requested by the current formatter.
        Device-wide values (like temperature and utilization) are reported for each process.

        As the slowest operation is getting device handles, we cache it inside the instance attributes.
        Note that this does nothing for a single query to this class.

        PIDs, reported by nvidia-smi, are from the host namespace and may (most probably) not
        exist in the container namespace. Currently, we don't have a reliable and not overly hacky way of matching it to
        a PID inside the container: we circumwent this problem in the `process_table`.
        """
        formatter = formatter or self.formatter
        device_table, device_process_table = ResourceTable(), ResourceTable()

        for device_id, handle in self.device_handles.items():
            common_info = {Resource.DEVICE_ID : device_id,
                           Resource.DEVICE_NAME : nvidia_smi.nvmlDeviceGetName(handle).decode()}

            # Inseparable device information like memory, temperature, power, etc. Request it only if needed
            if (formatter.get(Resource.DEVICE_UTIL, False) or
                formatter.get(Resource.DEVICE_UTIL_MA, False)):
                utilization = nvidia_smi.nvmlDeviceGetUtilizationRates(handle)
                common_info[Resource.DEVICE_UTIL] = utilization.gpu
                common_info[Resource.DEVICE_MEMORY_UTIL] = utilization.memory

                # Store values over requests to compute moving average of device utilization
                lst = self.device_utils[device_id]
                lst.append(utilization.gpu)
                common_info[Resource.DEVICE_UTIL_MA] = lst.get_average(size=window)

            if formatter.get(Resource.DEVICE_TEMP, False):
                temperature = nvidia_smi.nvmlDeviceGetTemperature(handle, nvidia_smi.NVML_TEMPERATURE_GPU)
                common_info[Resource.DEVICE_TEMP] = temperature

            if formatter.get(Resource.DEVICE_FAN, False):
                fan_speed = nvidia_smi.nvmlDeviceGetFanSpeed(handle)
                common_info[Resource.DEVICE_FAN] = fan_speed

            if formatter.get(Resource.DEVICE_POWER_USED, False):
                power_used = nvidia_smi.nvmlDeviceGetPowerUsage(handle)
                power_total = nvidia_smi.nvmlDeviceGetEnforcedPowerLimit(handle)

                common_info[Resource.DEVICE_POWER_USED] = power_used
                common_info[Resource.DEVICE_POWER_TOTAL] = power_total

            if (formatter.get(Resource.DEVICE_MEMORY_USED, False) or
                formatter.get(Resource.DEVICE_PROCESS_MEMORY_USED, False)):
                memory = nvidia_smi.nvmlDeviceGetMemoryInfo(handle)
                common_info[Resource.DEVICE_MEMORY_USED] = memory.used
                common_info[Resource.DEVICE_MEMORY_TOTAL] = memory.total

            # Collect individual processes info, if needed. Save it to both tables: in one as list, in other separately
            device_info = {**common_info}
            processes = nvidia_smi.nvmlDeviceGetComputeRunningProcesses(handle)
            device_info.update({Resource.DEVICE_PROCESS_N : 0,
                                Resource.DEVICE_PROCESS_PID : [],
                                Resource.DEVICE_PROCESS_MEMORY_USED : []})

            if processes:
                for process in processes:
                    pid, process_memory = process.pid, process.usedGpuMemory

                    # Update the aggregate device info table
                    device_info[Resource.DEVICE_PROCESS_N] += 1
                    device_info[Resource.DEVICE_PROCESS_PID].append(pid)
                    device_info[Resource.DEVICE_PROCESS_MEMORY_USED].append(process_memory)

                    # Update the table with individual processes
                    device_process_info = {**common_info}
                    device_process_info[Resource.DEVICE_PROCESS_PID] = pid
                    device_process_info[Resource.DEVICE_PROCESS_MEMORY_USED] = process_memory
                    device_process_table.append(device_process_info)

            device_table.append(device_info)
        return device_table, device_process_table

    def get_notebook_table(self, formatter=None):
        """ Collect information about all running Jupyter Notebooks inside all of the Jupyter Servers.
        Works with both v2 and v3 APIs.

        The most valuable information from this table is the mapping from `kernel_id` to `path` and `name`: all of
        other properties of a process can be retrieved by looking at the process (see `get_process_table`).

        TODO: once VSCode has stable standard and doc for ipykernel launches, add its parsing here.
        """
        #pylint: disable=import-outside-toplevel, missing-timeout
        servers = []
        try:
            from notebook.notebookapp import list_running_servers as list_running_servers_v2
            servers.extend(list(list_running_servers_v2()))
        except ImportError:
            pass
        try:
            from jupyter_server.serverapp import list_running_servers as list_running_servers_v3
            servers.extend(list(list_running_servers_v3()))
        except ImportError:
            pass

        _ = formatter # currently, not used

        # Information about all running kernels for all running servers
        notebook_table = ResourceTable()
        for server in servers:
            root_dir = server.get('root_dir') or server.get('notebook_dir') # for v2 and v3
            response = requests.get(requests.compat.urljoin(server['url'], 'api/sessions'),
                                    params={'token': server.get('token', '')})

            for instance in json.loads(response.text):
                kernel_id = instance['kernel']['id']
                notebook_info = {
                    Resource.NAME : instance['notebook']['name'],
                    Resource.PATH : os.path.join(root_dir, instance['notebook']['path']),
                    Resource.KERNEL : kernel_id,
                }
                notebook_table.append(notebook_info)
        return notebook_table

    def get_python_pids(self):
        """ PIDs of processes, which have `python` in its name. """
        python_pids = set()
        for pid in psutil.pids():
            name = pid_to_name(pid)
            if 'python' in name:
                python_pids.add(pid)
        return python_pids

    def get_process_table(self, formatter=None):
        """ Collect information about all Python processes.
        Information varies from properties of a process (its path, PID, NGID, status, etc) to system resource usage like
        CPU utilization or RSS. The table also includes some inferred columns like type and kernel_id.

        Some of the fields are intentionally left blank / with meaningless defaults: those are supposed to be filled by
        later merges / updates of the multiple tables. For example, processes that contain `kernel_id` in the name are
        matched on `kernel_id` with the result of `get_notebook_table` to fill in correct names and paths.

        If those fields are not updated with correct info, they would break table formatting: mainly, sorting and
        filtering. This is a good thing, as such occasions signal about something very wrong and unexpected.
        An example of this is an abandoned Jupyter Notebook, not managed by its Jupyter Server, or a host process, for
        some reasons visible inside the container.

        As `NBStat` can be run inside the container which has different namespace to the host, we are trying to match
        PIDs of the processes to the ones on the host. That is what the `NGID` column for: later we use
        either PIDs or NGIDs (whichever matches the DEVICE_PIDs) to merge with device information.
        This fallback should be easy to extend once we find new ways of inferring the host PID of a process.
        """
        formatter = formatter or self.formatter

        python_pids = self.get_python_pids()

        process_table = ResourceTable()
        for pid in python_pids:
            try:
                process = psutil.Process(pid)
                pid = process.pid

                with process.oneshot():
                    # Command used to start the Python interpreter
                    cmdline = ' '.join(process.cmdline())

                    # cwd with a default: access can be denied to current user
                    try:
                        cwd = process.cwd()
                    except psutil.AccessDenied:
                        cwd = ''

                    # Determine the type, name and path of the python process
                    kernel_id = KERNEL_ID_SEARCHER(cmdline)
                    vscode_key = VSCODE_KEY_SEARCHER(cmdline)
                    script_name = SCRIPT_NAME_SEARCHER(cmdline)

                    if kernel_id:
                        # The name will be changed by data from `notebook_table`.
                        # If not, then something very fishy is going on.
                        type_ = 'notebook'
                        name = kernel_id.group(1).split('-')[0] + '.ipynb'
                        path = os.path.join(cwd, name)
                        kernel_id = kernel_id.group(1)
                    elif vscode_key:
                        # Can't tell much more for processes run by VSCode for now
                        type_ = 'vscode'
                        name = vscode_key.group(1).split('-')[0] + '.ipynb'
                        path = kernel_id = vscode_key.group(1)
                    elif script_name:
                        type_ = 'script'
                        name = script_name.group(1) + '.py'
                        path = os.path.join(cwd, name)
                        kernel_id = None
                    else:
                        type_ = 'unknown'
                        name = 'unknown'
                        path = cwd
                        kernel_id = None

                    # PYTHON_PPID = PPID if parent is Python process else -1
                    ppid = process.ppid()
                    if ppid in python_pids:
                        # Spawned by one of other Python processes
                        type_ = 'subprocess'
                        python_ppid = ppid
                    elif 'containerd' in pid_to_name(ppid):
                        # Something very wrong is going on
                        type_ = 'containerd'
                        python_ppid = ppid
                    else:
                        # Spawned by non-Python process: terminal / Jupyter Server
                        python_ppid = -1

                    # Fill in the basic info
                    process_info = {
                        Resource.NAME : name,
                        Resource.PATH : path,
                        Resource.CMDLINE : cmdline,
                        Resource.TYPE : type_,
                        Resource.PID : pid,
                        Resource.PPID : ppid,
                        Resource.NGID : pid_to_ngid(pid),
                        Resource.PYTHON_PPID : python_ppid,
                        Resource.CREATE_TIME : process.create_time(),
                        Resource.KERNEL : kernel_id,
                        Resource.STATUS : process.status(),
                        Resource.PROCESS : process
                    }

                    # Gather resource info
                    if formatter.get(Resource.CPU, False):
                        process_info[Resource.CPU] = process.cpu_percent()

                    if formatter.get(Resource.RSS, False):
                        memory = process.memory_info()
                        process_info[Resource.RSS] = memory.rss

                process_table.append(process_info)

            except (psutil.AccessDenied, psutil.ZombieProcess, psutil.NoSuchProcess, FileNotFoundError):
                continue

        # Postprocess the table: update some entries with info from the others
        for entry in process_table:
            if entry[Resource.NAME] == 'unknown' and 'multiprocess' in entry[Resource.CMDLINE]:
                ppid = entry[Resource.PPID]

                for entry_ in process_table:
                    if entry_[Resource.PID] == ppid and entry_[Resource.PYTHON_PPID] == -1:
                        entry.update({key : entry_[key] for key in [Resource.NAME, Resource.PATH, Resource.KERNEL]})

        return process_table


    # Aggregate multiple ResourceTables into more representative tables
    def make_nbstat_table(self, formatter=None, sort=True, verbose=0, window=20):
        """ Prepare a `nbstat` view: a table, indexed by script/notebook name, with info about each of its processes.

        Parameters
        ----------
        sort : bool
            If True, then the table is sorted in the following order:
                - notebook name, which is used as table index
                - for each notebook, we display its process as the first row
                - then all device processes, sorted by device ID
                - then all other processes, sorted by create time
        verbose : {0, 1, 2}
            Sets the filtration of the table.
            If 0, then we keep only notebooks, which use at least one device. For them we keep only device processes.
            If 1, then we keep only notebooks, which use at least one device. For them we keep all processes.
            If 2, then we keep all notebooks and all processes for them.
        """
        # Collect all the data
        _, device_process_table = self.get_device_table(formatter=formatter, window=window) # ~20% of the time taken
        notebook_table = self.get_notebook_table(formatter=formatter)                       # ~15% of the time taken
        process_table = self.get_process_table(formatter=formatter)                         # ~45% of the time taken

        # Try to match the process pids (local namespace) to device pids (host). Merge on those
        table = process_table
        if device_process_table:
            device_pids = device_process_table[Resource.DEVICE_PROCESS_PID]

            def select_pid(entry):
                pid = entry[Resource.PID]
                ngid = entry[Resource.NGID]
                result = ngid if ngid in device_pids else pid
                return result

            table.add_column(Resource.HOST_PID, select_pid)
            self.nbstat_check_device_pids(device_pids, table, add_to_table=True)

            table = ResourceTable.merge(table, device_process_table,
                                        self_key=Resource.HOST_PID, other_key=Resource.DEVICE_PROCESS_PID)

        # Update entries: change `path` and `name` for Notebook from placeholders to proper ones
        if notebook_table:
            table.update(notebook_table, self_key=Resource.KERNEL, other_key=Resource.KERNEL, inplace=True)

        # Custom sort for processes: parent -> device processes -> non-device processes -> create time
        if sort:
            is_parent = lambda entry: entry[Resource.PYTHON_PPID] == -1
            table.add_column(Resource.IS_PARENT, is_parent)

            table.sort(key=[Resource.IS_PARENT, Resource.DEVICE_ID, Resource.CREATE_TIME],
                       reverse=[True, False, False], default=[0.0, 999, 0.0])

        # Filter non-device processes
        if verbose == 0:
            function = lambda entry: (entry.get(Resource.DEVICE_ID) is not None or entry[Resource.PYTHON_PPID] == -1)
            table.filter(function, inplace=True)

        # Sort index on create time
        table.set_index(Resource.PATH, inplace=True)
        if sort:
            uses_device = lambda entry: entry.get(Resource.DEVICE_ID) is not None
            table.add_column(Resource.USES_DEVICE, uses_device)
            table.sort_by_index(key=(Resource.USES_DEVICE, Resource.CREATE_TIME),
                                reverse=[True, False], aggregation=[max, min])

        # Filter non-device notebooks
        if verbose <= 1:
            function = lambda entry: (entry.get(Resource.DEVICE_ID) is not None)
            table.filter_by_index(function, inplace=True)
        return table

    def make_devicestat_table(self, formatter=None, window=20):
        """ A transposed `nbstat` view: the same information, but indexed with device ids. """
        device_table, device_process_table = self.get_device_table(formatter=formatter, window=window)
        notebook_table = self.get_notebook_table(formatter=formatter)
        process_table = self.get_process_table(formatter=formatter)

        # Try to match the process pids (local namespace) to device pids (host). Merge on those
        device_pids = device_process_table[Resource.DEVICE_PROCESS_PID]
        def select_pid(entry):
            pid = entry[Resource.PID]
            ngid = entry[Resource.NGID]
            result = ngid if ngid in device_pids else pid
            return result
        process_table.add_column(Resource.HOST_PID, select_pid)

        table = device_table.unroll(inplace=False)
        if process_table:
            table = table.merge(process_table, self_key=Resource.DEVICE_PROCESS_PID, other_key=Resource.HOST_PID)

        # Update entries: change `path` and `name` for Notebook from placeholders to proper ones
        if notebook_table:
            table.update(notebook_table, self_key=Resource.KERNEL, other_key=Resource.KERNEL, inplace=True)

        self.devicestat_check_device_pids(table)

        # A simple sort of entries and index
        table.sort(key=Resource.CREATE_TIME, reverse=False)
        table.set_index(Resource.DEVICE_ID, inplace=True)
        table.sort_by_index(key=Resource.DEVICE_ID, aggregation=min)
        return table

    def make_gpustat_table(self, formatter=None, window=20):
        """ A device-only view. Same information, as vanilla `gpustat`. """
        device_table, _ = self.get_device_table(formatter=formatter, window=window)
        device_table.set_index(Resource.DEVICE_ID)
        return device_table


    # Check for information consistency in multiple tables
    def nbstat_check_device_pids(self, device_pids, table, add_to_table=True):
        """ Check if some of `device pids` are not referenced in the `table`.
        Add them with template names and values, if needed.
        """
        set_device_pids = set(device_pids)
        set_host_pids = set(table[Resource.HOST_PID])
        if None in set_device_pids:
            set_device_pids.pop(None)
        if None in set_host_pids:
            set_host_pids.pop(None)

        if set_device_pids != set_host_pids:
            missing_pids = set_device_pids.difference(set_host_pids)
            self.warnings['missing_device_pids'] = missing_pids

            if add_to_table:
                entry_template = {key : None for key in table.columns or []}
                for missing_pid in sorted(missing_pids):
                    name = 'non-python' if psutil.pid_exists(missing_pid) else 'device_zombie'

                    entry = {
                        **entry_template,
                        Resource.NAME : name,
                        Resource.TYPE : name,
                        Resource.PATH : name,
                        Resource.STATUS : 'sleeping',
                        Resource.PID : missing_pid,
                        Resource.PPID : missing_pid,
                        Resource.NGID : missing_pid,
                        Resource.HOST_PID : missing_pid,
                        Resource.PYTHON_PPID : missing_pid,
                        Resource.CREATE_TIME : missing_pid, # for sort on `CREATE_TIME`
                    }
                    table.append(entry)

    def devicestat_check_device_pids(self, table):
        """ Check if some of the `device pids` are not matched to any Python processes.
        Add template names to them insted of empty ones.
        """
        self.warnings['missing_device_pids'] = set()
        for entry in table:
            if entry[Resource.DEVICE_PROCESS_PID] is not None and entry[Resource.HOST_PID] is None:
                missing_pid = entry[Resource.DEVICE_PROCESS_PID]
                self.warnings['missing_device_pids'].add(missing_pid)

                name = 'non-python' if psutil.pid_exists(missing_pid) else 'device_zombie'
                entry.update({Resource.NAME : name,
                              Resource.TYPE : name,
                              Resource.PATH : name,
                              Resource.STATUS : 'sleeping'})

    # Make formatted visualization of tables
    def get_view(self, name='nbstat', formatter=None, index_condition=None, force_styling=True,
                 sort=True, verbose=0, window=20, interval=None,
                 add_supheader=True, underline_supheader=True, bold_supheader=True, separate_supheader=False,
                 add_header=True, underline_header=True, bold_header=False, separate_header=True,
                 add_footnote=False, underline_footnote=False, bold_footnote=False, separate_footnote=True,
                 separate_index=True, separator='—', hide_similar=True,
                 process_memory_format='GB', device_memory_format='MB'):
        """ Get the desired view. Format it into colored string.
        Optionally, add a supheader (driver and CUDA info) and a footnote (total CPU / RSS usage) to the visualization.
        """
        formatter = formatter or self.formatter

        # Get the table
        if name.startswith('nb'):
            table = self.make_nbstat_table(formatter=formatter, sort=sort, verbose=verbose, window=window)
        elif name.startswith('device'):
            table = self.make_devicestat_table(formatter=formatter, window=window)
        elif name.startswith('gpu'):
            table = self.make_gpustat_table(formatter=formatter, window=window)
        else:
            raise ValueError('Wrong name of view to get!')

        # Filter index of the table by a regular expression
        if table and index_condition is not None:
            function = lambda index_value, _: bool(re.search(index_condition, str(index_value)))
            table.filter_on_index(function, inplace=True)

        # Create terminal instance
        terminal = self.make_terminal(force_styling=force_styling, separator=separator)

        # Make formatted strings
        lines = table.format(terminal=terminal, formatter=formatter,
                             add_header=add_header, underline_header=underline_header,
                             bold_header=bold_header, separate_header=separate_header,
                             separate_index=separate_index, hide_similar=hide_similar,
                             process_memory_format=process_memory_format, device_memory_format=device_memory_format)

        if add_supheader:
            lines = self.add_supheader(lines, terminal=terminal, interval=interval,
                                       underline=underline_supheader, bold=bold_supheader, separate=separate_supheader)
        if add_footnote:
            lines = self.add_footnote(lines, terminal=terminal,
                                      underline=underline_footnote, bold=bold_footnote, separate=separate_footnote,
                                      process_memory_format=process_memory_format)

        # Placeholder for empty table
        if not table:
            width = terminal.length(lines[-1])
            placeholder = terminal.rjust(terminal.bold + 'No entries to display!' + terminal.normal, width)
            lines.insert(2, placeholder)

        # For debug purposes
        self._table = table
        return '\n'.join(lines)


    def make_terminal(self, force_styling, separator):
        """ Create terminal instance. """
        terminal = Terminal(kind=os.getenv('TERM'), force_styling=force_styling if force_styling else None)
        terminal.separator_symbol = terminal.bold + separator + terminal.normal
        terminal._normal = '\x1b[0;10m' # pylint: disable=protected-access

        # Change some methods to a faster versions
        terminal.length = true_len
        terminal.rjust = true_rjust
        terminal.center = true_center
        return terminal

    def add_line(self, lines, parts, terminal, position, separator_position, underline, bold):
        """ Add line, created from joined `parts`, to `lines`, in desired `position`. """
        parts = [part for part in parts if part]
        if underline:
            parts = [terminal.underline + part for part in parts]
        if bold:
            parts = [terminal.bold + part for part in parts]
        parts = [part + terminal.normal for part in parts]
        added_line = '    '.join(parts)

        added_line_width = terminal.length(added_line)
        table_width = terminal.length(lines[0])

        if added_line_width <= table_width:
            added_line = terminal.rjust(added_line, table_width)
        else:
            lines = [terminal.rjust(line, added_line_width) for line in lines]
        lines.insert(position, added_line)

        if separator_position is not None:
            lines.insert(separator_position, terminal.separator_symbol * terminal.length(added_line))
        return lines

    def add_supheader(self, lines, terminal, interval=None, underline=True, bold=True, separate=True):
        """ Add a supheader with info about current time, driver and CUDA versions. """
        timestamp = time.strftime('%Y-%m-%d %H:%M:%S')
        interval_info = f'Interval: {interval:2.1f}s' if interval is not None else ''
        driver_version = '.'.join(nvidia_smi.nvmlSystemGetDriverVersion().decode().split('.')[:-1])
        cuda_version = nvidia_smi.nvmlSystemGetNVMLVersion().decode()[:4]

        parts = [
            timestamp,
            f'Driver Version: {driver_version}',
            f'CUDA Version: {cuda_version}',
            interval_info,
        ]
        lines = self.add_line(lines=lines, parts=parts, terminal=terminal,
                              position=0, separator_position=1 if separate else None,
                              underline=underline, bold=bold)
        return lines

    def add_footnote(self, lines, terminal, underline=True, bold=True, separate=True, process_memory_format='GB'):
        """ Add a footnote with info about current CPU and RSS usage. """
        vm = psutil.virtual_memory()
        vm_used, unit = format_memory(vm.used, process_memory_format)
        vm_total, unit = format_memory(vm.total, process_memory_format)
        n_digits = len(str(vm_total))

        parts = [
            f'{terminal.bold + terminal.cyan}SYSTEM CPU: {psutil.cpu_percent():6}%',
            f'{terminal.bold + terminal.cyan}SYSTEM RSS: {vm_used:>{n_digits}} / {vm_total} {unit}'
        ]

        lines = self.add_line(lines=lines, parts=parts, terminal=terminal,
                              position=len(lines), separator_position=None,
                              underline=underline, bold=bold)
        if separate:
            lines.insert(-1, ' ')
        return lines
